/* RTcmix  - Copyright (C) 2004  The RTcmix Development Team
   See ``AUTHORS'' for a list of contributors. See ``LICENSE'' for
   the license to this software and for a DISCLAIMER OF ALL WARRANTIES.
*/
/* written by John Gibson, based on the classic cmix Minc by Lars Graf */

#ifndef _MINC_INTERNAL_H_
#define _MINC_INTERNAL_H_ 1

#include <float.h>   /* for epsilon constants */
#include <maxdispargs.h>
#include <ugens.h>
#include "rename.h"
#include "minc_defs.h"
#include "utils.h"
#include "RefCounted.h"
#include <vector>

#ifdef DEBUG
   #define DPRINT(...) rtcmix_print(__VA_ARGS__)
#else
   #define DPRINT(...)
#endif

/* important Minc tuning parameters */
#define YYLMAX   2048      /* maximum yacc line length */
#define MAXSTACK 15        /* depth of function call or list recursion */
#define HASHSIZE 107       /* number of buckets in string table */

enum MincWarningLevel {
    MincNoWarnings = 0,
    MincNoDefaultedArgWarnings = 1,
    MincAllWarnings = 2,
    MincWarningsAsErrors = 5
};

typedef enum {
	OpZero = 0,
	OpFree,
	OpPlus,
	OpMinus,
	OpMul,
	OpDiv,
	OpMod,
	OpPow,
	OpNeg,
	OpEqual,
	OpNotEqual,
	OpLess,
	OpGreater,
	OpLessEqual,
	OpGreaterEqual,
    OpPlusPlus,
    OpMinusMinus
} OpKind;

/* prototypes for internal Minc use */

#define EPSILON DBL_EPSILON

/* error.cpp */
void sys_error(const char *msg);
void minc_advise(const char *msg, ...);
void minc_warn(const char *msg, ...);
void minc_die(const char *msg, ...);
void minc_internal_error(const char *msg, ...);
extern "C" void yyerror(const char *msg);

class RTException
{
public:
	RTException(const char *msg) : _mesg(msg) {}
	const char *mesg() { return _mesg; }
private:
	const char *_mesg;
};

class RTFatalException : public RTException
{
public:
	RTFatalException(const char *msg) : RTException(msg) {}
};

// Such as attempting to multiply two MincStrings

class UnsupportedOperationException : public RTException
{
public:
	UnsupportedOperationException(const char *msg) : RTException(msg) {}
};

class InvalidOperatorException : public RTException
{
public:
	InvalidOperatorException(const char *msg) : RTException(msg) {}
};

class NonmatchingTypeException : public RTException
{
public:
	NonmatchingTypeException(const char *msg) : RTException(msg) {}
};

// Such as indexing a MincList with a MincString

class InvalidTypeException : public RTFatalException
{
public:
	InvalidTypeException(const char *msg) : RTFatalException(msg) {}
};

class UndeclaredVariableException : public RTFatalException
{
public:
	UndeclaredVariableException(const char *msg) : RTFatalException(msg) {}
};

class ReclaredVariableException : public RTFatalException
{
public:
	ReclaredVariableException(const char *msg) : RTFatalException(msg) {}
};

// Such as divide or mod by zero

class ArithmaticException : public RTFatalException
{
public:
    ArithmaticException(const char *msg) : RTFatalException(msg) {}
};

class MincObject
{
public:
    void *operator new(size_t size);
    void operator delete(void *);
};

typedef double MincFloat;
typedef const char *MincString;
typedef void *MincHandle;  // contents of this is opaque to Minc

enum MincDataType {
    MincVoidType = 0,
    MincFloatType = 1,       /* a floating point number, either float or double */
    MincStringType = 2,
    MincHandleType = 4,
    MincListType = 8,
    MincStructType = 16
};

class MincValue;
class MincList;
class Node;

union YYSTYPE {
    int ival;
    Node *node;
    char *str;
};
#define YYSTYPE_IS_DECLARED   /* keep bison from declaring YYSTYPE as an int */

enum LookupType { AnyLevel = 0, GlobalLevel = 1, ThisLevel = 2 };

void printargs(const char *funcname, const Arg arglist[], const int nargs);

char *strsave(const char *str);
void clear_elem(MincValue *);
void unref_value_list(MincValue *);

/* utils.cpp */
int is_float_list(const MincList *list);
MincFloat *float_list_to_array(const MincList *list);
MincList *array_to_float_list(const MincFloat *array, const int len);
const char *MincTypeName(MincDataType type);
void increment_score_line_offset(int offset);
int get_score_line_offset();

int hash(const char *c);
int cmp(MincFloat f1, MincFloat f2);

char *emalloc(long nbytes);
void efree(void *mem);

inline void *	MincObject::operator new(size_t size)
{
	return emalloc(size);
}

inline void	MincObject::operator delete(void *ptr)
{
	efree(ptr);
}

#endif /* _MINC_INTERNAL_H_ */
